@@ -1,26 +1,14 @@
 import re
 
-
-re_spaces = re.compile(ur"\s+")
+from django.utils import six
 
 
-class cached_property(object):
-    """
-    Decorator that creates converts a method with a single
-    self argument into a property cached on the instance.
-    (not available in django 1.2)
-    """
-    def __init__(self, func):
-        self.func = func
-
-    def __get__(self, instance, type):
-        res = instance.__dict__[self.func.__name__] = self.func(instance)
-        return res
+re_spaces = re.compile(r"\s+")
 
 
 def combine_css_classes(classes, new_classes):
     if not classes:
-        if isinstance(new_classes, basestring):
+        if isinstance(new_classes, six.string_types):
             return new_classes
         else:
             try:
@@ -28,7 +16,7 @@ def combine_css_classes(classes, new_classes):
             except TypeError:
                 return new_classes
 
-    if isinstance(classes, basestring):
+    if isinstance(classes, six.string_types):
         classes = set(re_spaces.split(classes))
     else:
         try:
@@ -36,7 +24,7 @@ def combine_css_classes(classes, new_classes):
         except TypeError:
             return classes
 
-    if isinstance(new_classes, basestring):
+    if isinstance(new_classes, six.string_types):
         new_classes = set(re_spaces.split(new_classes))
     else:
         try:
